from __future__ import annotations

import json
import os
from datetime import datetime
from functools import partial
from typing import Any, Awaitable, Callable, Dict, List, Optional, Tuple, Union

from anyio import TASK_STATUS_IGNORED, Lock, create_task_group, sleep
from anyio.abc import TaskStatus
from fastapi import APIRouter, Cookie, Depends, HTTPException, Request, WebSocket, status
from fastapi.responses import RedirectResponse
from httpx import AsyncClient
from jupyterhub.services.auth import HubOAuth
from jupyterhub.utils import isoformat
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine
from sqlalchemy.future import select
from typing_extensions import Annotated

from jupyverse_api import Router
from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User

from .config import AuthJupyterHubConfig
from .db import Base, UserDB
from .models import JupyterHubUser


def auth_factory(
    app: App,
    auth_jupyterhub_config: AuthJupyterHubConfig,
):
    class AuthJupyterHub(Auth, Router):
        def __init__(self) -> None:
            super().__init__(app)
            self.db_engine = create_async_engine(auth_jupyterhub_config.db_url)
            self.db_session = AsyncSession(self.db_engine)

            self.http_client = AsyncClient()
            self.hub_auth = HubOAuth()
            self.db_lock = Lock()
            self.activity_url = os.environ.get("JUPYTERHUB_ACTIVITY_URL")
            self.server_name = os.environ.get("JUPYTERHUB_SERVER_NAME")
            self.http_client = AsyncClient()

            router = APIRouter()

            @router.get("/oauth_callback")
            async def get_oauth_callback(
                request: Request,
                code: str | None = None,
                state: str | None = None,
            ):
                if code is None:
                    raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)

                cookie_state = request.cookies.get(self.hub_auth.state_cookie_name)
                if state is None or state != cookie_state:
                    raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)

                token = self.hub_auth.token_for_code(code)
                hub_user = await self.hub_auth.user_for_token(token, use_cache=False, sync=False)
                async with self.db_lock:
                    self.db_session.add(
                        UserDB(
                            token=token,
                            name=hub_user["name"],
                            username=hub_user["name"],
                        ),
                    )
                    await self.db_session.commit()

                next_url = self.hub_auth.get_next_url(cookie_state)
                response = RedirectResponse(next_url)
                response.set_cookie(key="jupyverse_jupyterhub_token", value=token)
                return response

            @router.get("/api/me")
            async def get_api_me(
                request: Request,
                user: User = Depends(self.current_user()),
            ):
                checked_permissions: Dict[str, List[str]] = {}
                permissions = json.loads(
                    dict(request.query_params).get("permissions", "{}").replace("'", '"')
                )
                if permissions:
                    user_permissions: Dict[str, List[str]] = {}
                    for resource, actions in permissions.items():
                        user_resource_permissions = user_permissions.get(resource, [])
                        allowed = checked_permissions[resource] = []
                        for action in actions:
                            if action in user_resource_permissions:
                                allowed.append(action)

                keys = ["username", "name", "display_name", "initials", "avatar_url", "color"]
                identity = {k: getattr(user, k) for k in keys}
                return {
                    "identity": identity,
                    "permissions": checked_permissions,
                }

            self.include_router(router)

        def current_user(self, permissions: Optional[Dict[str, List[str]]] = None) -> Callable:
            async def _(
                request: Request,
                jupyverse_jupyterhub_token: Annotated[Union[str, None], Cookie()] = None,
            ):
                if jupyverse_jupyterhub_token is not None:
                    hub_user = await self.hub_auth.user_for_token(
                        jupyverse_jupyterhub_token, use_cache=False, sync=False
                    )
                    scopes = self.hub_auth.check_scopes(self.hub_auth.access_scopes, hub_user)
                    if not scopes:
                        raise HTTPException(
                            status_code=status.HTTP_403_FORBIDDEN,
                            detail=f"User {hub_user['name']} cannot access this server",
                        )

                    async with self.db_lock:
                        user_db = await self.db_session.scalar(
                            select(UserDB).filter_by(token=jupyverse_jupyterhub_token)
                        )
                    user = JupyterHubUser.model_validate(user_db)
                    if self.activity_url:
                        headers = {
                            "Authorization": f"token {self.hub_auth.api_token}",
                            "Content-Type": "application/json",
                        }
                        last_activity = isoformat(datetime.utcnow())
                        self.task_group.start_soon(
                            partial(
                                self.http_client.post,
                                self.activity_url,
                                headers=headers,
                                json={
                                    "servers": {self.server_name: {"last_activity": last_activity}}
                                },
                            )
                        )
                    return user

                if permissions:
                    raise HTTPException(
                        status_code=status.HTTP_403_FORBIDDEN,
                    )

                state = self.hub_auth.generate_state(next_url=str(request.url))
                raise HTTPException(
                    status_code=status.HTTP_307_TEMPORARY_REDIRECT,
                    headers={
                        "Location": f"{self.hub_auth.login_url}&state={state}",
                        "Set-Cookie": f"{self.hub_auth.state_cookie_name}={state}",
                    },
                )

            return _

        async def start(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED) -> None:
            async with self.db_engine.begin() as conn:
                await conn.run_sync(Base.metadata.create_all)

            async with create_task_group() as tg:
                self.task_group = tg
                task_status.started()
                await sleep(float("inf"))

        async def stop(self) -> None:
            await self.http_client.aclose()
            await self.db_session.close()
            await self.db_engine.dispose()

        async def update_user(
            self, jupyverse_jupyterhub_token: Annotated[Union[str, None], Cookie()] = None
        ) -> Callable:
            async def _(data: Dict[str, Any]) -> JupyterHubUser | None:
                if jupyverse_jupyterhub_token is not None:
                    async with self.db_lock:
                        user_db = await self.db_session.scalar(
                            select(UserDB).filter_by(token=jupyverse_jupyterhub_token)
                        )
                        for k, v in data.items():
                            setattr(user_db, k, v)
                        await self.db_session.commit()
                    user = JupyterHubUser.model_validate(user_db)
                    return user
                return None

            return _

        def websocket_auth(
            self,
            permissions: Optional[Dict[str, List[str]]] = None,
        ) -> Callable[[Any], Awaitable[Optional[Tuple[Any, Optional[Dict[str, List[str]]]]]]]:
            async def _(
                websocket: WebSocket,
            ) -> Optional[Tuple[Any, Optional[Dict[str, List[str]]]]]:
                accept_websocket = False
                if "jupyverse_jupyterhub_token" in websocket._cookies:
                    jupyverse_jupyterhub_token = websocket._cookies["jupyverse_jupyterhub_token"]
                    async with self.db_lock:
                        user_db = await self.db_session.scalar(
                            select(UserDB).filter_by(token=jupyverse_jupyterhub_token)
                        )
                    if user_db:
                        accept_websocket = True
                if accept_websocket:
                    return websocket, permissions
                else:
                    await websocket.close(code=status.WS_1008_POLICY_VIOLATION)
                    return None

            return _

    return AuthJupyterHub()



from anyio import create_task_group
from fps import Module

from jupyverse_api.app import App
from jupyverse_api.auth import Auth, AuthConfig

from .config import AuthJupyterHubConfig
from .routes import auth_factory


class AuthJupyterHubModule(Module):
    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self.auth_jupyterhub_config = AuthJupyterHubConfig(**kwargs)

    async def prepare(self) -> None:
        self.put(self.auth_jupyterhub_config, AuthConfig)
        app = await self.get(App)

        self.auth_jupyterhub = auth_factory(app, self.auth_jupyterhub_config)

        async with create_task_group() as self.tg:
            await self.tg.start(self.auth_jupyterhub.start)
            self.put(self.auth_jupyterhub, Auth)
            self.done()

    async def stop(self) -> None:
        await self.auth_jupyterhub.stop()
        self.tg.cancel_scope.cancel()



from fps import Module

from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.contents import Contents

from .routes import _Contents


class ContentsModule(Module):
    async def prepare(self) -> None:
        app = await self.get(App)
        auth = await self.get(Auth)  # type: ignore

        contents = _Contents(app, auth)
        self.put(contents, Contents)



import pytest
from fps import get_root_module, merge_config
from httpx import AsyncClient
from httpx_ws import WebSocketUpgradeError, aconnect_ws
from utils import authenticate_client

from jupyverse_api.auth import AuthConfig

CONFIG = {
    "jupyverse": {
        "type": "jupyverse",
        "modules": {
            "app": {
                "type": "app",
            },
            "auth": {
                "type": "auth",
                "config": {
                    "test": True,
                },
            },
            "contents": {
                "type": "contents",
            },
            "frontend": {
                "type": "frontend",
            },
            "lab": {
                "type": "lab",
            },
            "jupyterlab": {
                "type": "jupyterlab",
            },
            "kernels": {
                "type": "kernels",
            },
            "yjs": {
                "type": "yjs",
            },
        }
    }
}


@pytest.mark.anyio
async def test_kernel_channels_unauthenticated(unused_tcp_port):
    config = merge_config(CONFIG, {"jupyverse": {"config": {"port": unused_tcp_port}}})
    async with get_root_module(config):
        with pytest.raises(WebSocketUpgradeError):
            async with aconnect_ws(
                f"http://127.0.0.1:{unused_tcp_port}/api/kernels/kernel_id_0/channels?session_id=session_id_0",
            ):
                pass


@pytest.mark.anyio
async def test_kernel_channels_authenticated(unused_tcp_port):
    config = merge_config(CONFIG, {"jupyverse": {"config": {"port": unused_tcp_port}}})
    async with get_root_module(config), AsyncClient() as http:
        await authenticate_client(http, unused_tcp_port)
        async with aconnect_ws(
            f"http://127.0.0.1:{unused_tcp_port}/api/kernels/kernel_id_0/channels?session_id=session_id_0",
            http,
        ):
            pass


@pytest.mark.anyio
@pytest.mark.parametrize("auth_mode", ("noauth", "token", "user"))
async def test_root_auth(auth_mode, unused_tcp_port):
    config = merge_config(
        CONFIG,
        {
            "jupyverse": {
                "config": {"port": unused_tcp_port},
                "modules": {
                    "auth": {
                        "config": {
                            "mode": auth_mode,
                        }
                    }
                }
            }
        }
    )
    async with get_root_module(config), AsyncClient() as http:
        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}/")
        if auth_mode == "noauth":
            expected = 302
        elif auth_mode in ["token", "user"]:
            expected = 403

        assert response.status_code == expected
        assert response.headers["content-type"] == "application/json"


@pytest.mark.anyio
@pytest.mark.parametrize("auth_mode", ("noauth",))
async def test_no_auth(auth_mode, unused_tcp_port):
    config = merge_config(
        CONFIG,
        {
            "jupyverse": {
                "config": {"port": unused_tcp_port},
                "modules": {
                    "auth": {
                        "config": {
                            "mode": auth_mode,
                        }
                    }
                }
            }
        }
    )
    async with get_root_module(config), AsyncClient() as http:
        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}/lab")
        assert response.status_code == 200


@pytest.mark.anyio
@pytest.mark.parametrize("auth_mode", ("token",))
async def test_token_auth(auth_mode, unused_tcp_port):
    config = merge_config(
        CONFIG,
        {
            "jupyverse": {
                "config": {"port": unused_tcp_port},
                "modules": {
                    "auth": {
                        "config": {
                            "mode": auth_mode,
                        }
                    }
                }
            }
        }
    )
    async with get_root_module(config) as jupyverse, AsyncClient() as http:
        auth_config = await jupyverse.get(AuthConfig)

        # no token provided, should not work
        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}/")
        assert response.status_code == 403
        # token provided, should work
        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}/?token={auth_config.token}")
        assert response.status_code == 302


@pytest.mark.anyio
@pytest.mark.parametrize("auth_mode", ("user",))
@pytest.mark.parametrize(
    "permissions",
    (
        {},
        {"admin": ["read"], "foo": ["bar", "baz"]},
    ),
)
async def test_permissions(auth_mode, permissions, unused_tcp_port):
    config = merge_config(
        CONFIG,
        {
            "jupyverse": {
                "config": {"port": unused_tcp_port},
                "modules": {
                    "auth": {
                        "config": {
                            "mode": auth_mode,
                        }
                    }
                }
            }
        }
    )
    async with get_root_module(config), AsyncClient() as http:
        await authenticate_client(http, unused_tcp_port, permissions=permissions)
        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}/auth/user/me")
        if "admin" in permissions.keys():
            # we have the permissions
            assert response.status_code == 200
        else:
            # we don't have the permissions
            assert response.status_code == 403



import structlog
from fps import Module

from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.jupyterlab import JupyterLabConfig
from jupyverse_api.lab import Lab
from jupyverse_api.notebook import Notebook

from .routes import _Notebook

logger = structlog.get_logger()


class NotebookModule(Module):
    async def prepare(self) -> None:
        self.put(None, JupyterLabConfig)
        app = await self.get(App)
        auth = await self.get(Auth)  # type: ignore[type-abstract]
        frontend_config = await self.get(FrontendConfig)
        lab = await self.get(Lab)  # type: ignore[type-abstract]

        notebook = _Notebook(app, auth, frontend_config, lab)
        self.put(notebook, Notebook)



import pytest
from anyio import create_task_group, sleep

from jupyverse_api import ResourceLock

pytestmark = pytest.mark.anyio


async def do_op(operation, resource_lock, operations):
    op, path = operation
    async with resource_lock(path):
        operations.append(operation + ["start"])
        await sleep(0.1)
        operations.append(operation + ["done"])


async def test_resource_lock():
    resource_lock = ResourceLock()

    # test concurrent accesses to the same resource
    idx = "idx"
    operations = []
    async with create_task_group() as tg:
        tg.start_soon(do_op, [0, idx], resource_lock, operations)
        await sleep(0.01)
        tg.start_soon(do_op, [1, idx], resource_lock, operations)

    assert operations == [
        [0, idx, "start"],
        [0, idx, "done"],
        [1, idx, "start"],
        [1, idx, "done"],
    ]

    # test concurrent accesses to different files
    idx0 = "idx0"
    idx1 = "idx1"
    operations = []
    async with create_task_group() as tg:
        tg.start_soon(do_op, [0, idx0], resource_lock, operations)
        await sleep(0.01)
        tg.start_soon(do_op, [1, idx1], resource_lock, operations)
        await sleep(0.01)
        tg.start_soon(do_op, [2, idx0], resource_lock, operations)
        await sleep(0.01)
        tg.start_soon(do_op, [3, idx1], resource_lock, operations)

    assert operations == [
        [0, idx0, "start"],
        [1, idx1, "start"],
        [0, idx0, "done"],
        [2, idx0, "start"],
        [1, idx1, "done"],
        [3, idx1, "start"],
        [2, idx0, "done"],
        [3, idx1, "done"],
    ]



from fps import Module

from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.resource_usage import ResourceUsage, ResourceUsageConfig

from .routes import _ResourceUsage


class ResourceUsageModule(Module):
    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self.resource_usage_config = ResourceUsageConfig(**kwargs)

    async def prepare(self) -> None:
        app = await self.get(App)
        auth = await self.get(Auth)  # type: ignore[type-abstract]

        resource_usage = _ResourceUsage(app, auth, self.resource_usage_config)
        self.put(resource_usage, ResourceUsage)



from fps import Module

from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.jupyterlab import JupyterLabConfig
from jupyverse_api.lab import Lab

from .routes import _Lab


class LabModule(Module):
    async def prepare(self) -> None:
        app = await self.get(App)
        auth = await self.get(Auth)  # type: ignore[type-abstract]
        frontend_config = await self.get(FrontendConfig)
        jupyterlab_config = await self.get(JupyterLabConfig)

        lab = _Lab(app, auth, frontend_config, jupyterlab_config)
        self.put(lab, Lab)



from typing import Any, Dict

from anyio import Event
from pydantic import BaseModel

from .app import App

__version__ = "0.7.5"


class Singleton(type):
    _instances: Dict = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Config(BaseModel):
    model_config = {"extra": "forbid"}


class Router:
    _app: App

    def __init__(
        self,
        app: App,
    ) -> None:
        self._app = app

    @property
    def _type(self):
        return self.__class__.__name__

    def include_router(self, router, **kwargs):
        self._app._include_router(router, self._type, **kwargs)

    def mount(self, path: str, *args, **kwargs) -> None:
        self._app._mount(path, self._type, *args, **kwargs)

    def add_middleware(self, middleware, *args, **kwargs) -> None:
        self._app.add_middleware(middleware, *args, **kwargs)


class ResourceLock:
    """ResourceLock ensures that accesses cannot be done concurrently on the same resource.
    """
    _locks: Dict[Any, Event]

    def __init__(self):
        self._locks = {}

    def __call__(self, idx: Any):
        return _ResourceLock(idx, self._locks)


class _ResourceLock:
    _idx: Any
    _locks: Dict[Any, Event]
    _lock: Event

    def __init__(self, idx: Any, locks: Dict[Any, Event]):
        self._idx = idx
        self._locks = locks

    async def __aenter__(self):
        while True:
            if self._idx not in self._locks:
                break
            await self._locks[self._idx].wait()
        self._locks[self._idx] = self._lock = Event()

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        self._lock.set()
        del self._locks[self._idx]
